using System;
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace MyDirect3DProject
{
	/// <summary>
	/// This is the main class of my Direct3D application
	/// </summary>
	public class MainClass : Form
	{
		/// <summary>
		/// The rendering device
		/// </summary>
		private Device device;
		private VertexBuffer vb;
		private IndexBuffer ib;
		private const int Rows = 42;
		private const int Columns = 53;
		private Material myMaterial;
		private Texture myTexture;

		private long frames;
		private System.DateTime startTime;

		public MainClass()
		{
			this.ClientSize = new System.Drawing.Size(640, 480);
			this.Text = "Direct3D Project";
			this.SetStyle(ControlStyles.AllPaintingInWmPaint|ControlStyles.Opaque, true);
		}
		
		public bool InitializeGraphics()
		{
			try {
				startTime = System.DateTime.Now;
				
				// Now let's setup the Direct3D stuff
				PresentParameters presentParams = new PresentParameters();
				presentParams.Windowed = true;
				presentParams.SwapEffect = SwapEffect.Discard;
				presentParams.EnableAutoDepthStencil = true;			
				presentParams.AutoDepthStencilFormat = DepthFormat.D16;
				
				// Create the device
				device = new Device(0, DeviceType.Hardware, this,
				                    CreateFlags.HardwareVertexProcessing, presentParams);
				
				// Setup the event handlers for the device
				device.DeviceLost     += new EventHandler(this.InvalidateDeviceObjects);
				device.DeviceReset    += new EventHandler(this.RestoreDeviceObjects);
				device.Disposing      += new EventHandler(this.DeleteDeviceObjects);
				device.DeviceResizing += new CancelEventHandler(this.EnvironmentResizing);
				
				vb = new VertexBuffer(typeof(CustomVertex.PositionNormalTextured),
				                                   (Rows+1)*(Columns+1), device, Usage.WriteOnly,
				                                   0, Pool.Managed);
				CustomVertex.PositionNormalTextured[] vtx = (CustomVertex.PositionNormalTextured[]) vb.Lock(0, LockFlags.None);
				for(int i = 0; i < (Rows+1)*(Columns+1); i++)
				{
					float x = 2.0f*(i % (Columns+1))/(float)Columns - 1.0f;
					float y = 2.0f*(i / (Columns+1))/(float)Rows - 1.0f;
					float z =  0.5f+0.2f*(float)Math.Sin(8.0*x);
					float nx = 0.2f*8.0f*(float)Math.Cos(8.0*x);
					float ny = 0.0f;
					float nz = -1.0f;
					float u = (i % (Columns+1))/(float)Columns;
					float v = (i / (Columns+1))/(float)Rows;
					vtx[i] = new CustomVertex.PositionNormalTextured(x, y, z, nx, ny, nz, u, v);
				}
				vb.Unlock();			
				
				ib = new IndexBuffer(typeof(short), 2*3*Rows*Columns, device, Usage.WriteOnly, Pool.Managed);
				short[] ind = (short[]) ib.Lock(0, LockFlags.None);
				for(int i = 0; i < Rows*Columns; i++) // zählt die Vierecke (= je zwei Dreiecke) durch
				{
					// Es gibt pro Zeile einen Vertex mehr (Columns + 1) als Vierecke!
					int iBase = (i % Columns) + (i / Columns)*(Columns + 1);
					ind[6*i  ] = (short)(iBase);
					ind[6*i+1] = (short)(iBase + Columns + 1);
					ind[6*i+2] = (short)(iBase + Columns + 2);
					ind[6*i+3] = (short)(iBase + Columns + 2);
					ind[6*i+4] = (short)(iBase + 1);
					ind[6*i+5] = (short)(iBase);					
				}
				ib.Unlock();
				
				myMaterial = new Material();
				myMaterial.Emissive = Color.FromArgb(23, 23, 33);
				myMaterial.Ambient = Color.FromArgb(23, 23, 33);
				myMaterial.Diffuse = Color.FromArgb(240, 255, 240);
				myMaterial.Specular = Color.White;
				myMaterial.SpecularSharpness = 100.0f;
				
				// zum Starten innerhalb von SharpDevelop: zwei Verzeichisse hochgehen
				myTexture = TextureLoader.FromFile(device, "../../rgba.tga"); // aus Direct3DX
				
				return true;
			} catch (DirectXException) {
				return false;
			}			
		}
		
		protected virtual void InvalidateDeviceObjects(object sender, EventArgs e)
		{
		}
		
		protected virtual void RestoreDeviceObjects(object sender, EventArgs e)
		{
		}
		
		protected virtual void DeleteDeviceObjects(object sender, EventArgs e)
		{
			// passiert hier auch alles automatisch;
			// manchmal ist es aber sinnvoller oder sogar nötig,
			// Objekte vor Programmende zu entsorgen
			ib.Dispose();
			vb.Dispose();	
			myTexture.Dispose();
		}
		
		protected virtual void EnvironmentResizing(object sender, CancelEventArgs e)
		{
		}
		
		/// <summary>
		/// This method moves the scene
		/// </summary>
		protected virtual void FrameMove()
		{
			// TODO : Frame movement
		}
		
		/// <summary>
		/// This method renders the scene
		/// </summary>
		protected virtual void Render()
		{
			if (device != null) {

				double time = (System.DateTime.Now - startTime).TotalSeconds;
				
				device.Clear(ClearFlags.Target|ClearFlags.ZBuffer, Color.Blue, 1.0f, 0);
				device.BeginScene();

				device.RenderState.CullMode = Cull.None;
				
				device.VertexFormat = CustomVertex.PositionNormalTextured.Format; // NICHT VERGESSEN!
				device.SetStreamSource(0, vb, 0);
				device.Indices = ib;
				
				device.Lights[0].Type = LightType.Directional;
				device.Lights[0].Diffuse = Color.Beige;
				device.Lights[0].Specular = Color.White;
				device.Lights[0].Direction = new Vector3(1.0f, -1.0f, 1.0f); // muss nicht Länge 1 haben
				device.Lights[0].Commit();
				device.Lights[0].Enabled = true;

				device.Material = myMaterial;
				device.RenderState.NormalizeNormals = true; // schneller ist, mit Normalen der Länge 1 zu arbeiten
				device.RenderState.Ambient = Color.DarkSlateBlue;
				device.RenderState.Lighting = true;
				device.RenderState.SpecularEnable = true;
				
				// Transparenz
				// device.RenderState.SourceBlend = Blend.SourceAlpha;				
				// device.RenderState.DestinationBlend = Blend.InvSourceAlpha;
				// device.RenderState.BlendOperation = BlendOperation.Add;
				// device.RenderState.AlphaBlendEnable = true;				
				
				device.SetTexture(0, myTexture);
				device.SamplerState[0].MagFilter = TextureFilter.Linear;
				device.SamplerState[0].MinFilter = TextureFilter.Linear;
				device.SamplerState[0].MipFilter = TextureFilter.Linear;			
				
				device.Transform.Projection = Matrix.PerspectiveFovLH(0.8f, Width/(float)Height, 0.1f, 50.0f);
				device.Transform.View = Matrix.LookAtLH(new Vector3(0.0f, 0.0f, -2.0f), new Vector3(), new Vector3(0.0f, 1.0f, 0.0f));
				device.Transform.World = Matrix.RotationX(1.2f)*Matrix.RotationY(0.8f*(float)time);

				device.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, (Rows+1)*(Columns+1), 0, 2*Rows*Columns);
				
				device.EndScene();
				device.Present();
				
				// unschön: irgendwann nach einer halben Ewigkeit läuft die Variable frames über
				frames ++;
				Text = (frames/time).ToString("###.0") + " fps";
			}
		}
		
		/// <summary>
		/// Our main loop
		/// </summary>
		public void Run()
		{
			// While the form is still valid, render and process messages
			while (Created) {
				FrameMove();	
				Render();
				Application.DoEvents();
			}
		}
		
		protected override void OnPaint(PaintEventArgs e)
		{
			this.Render();
		}
		
		protected override void OnKeyPress(KeyPressEventArgs e)
		{
			base.OnKeyPress(e);
			if ((int)e.KeyChar == (int)System.Windows.Forms.Keys.Escape) {
				this.Close(); 
			}
		}
		
		/// <summary>
		/// The main entry point for the application
		/// </summary>
		static void Main()
		{
			using (MainClass mainClass = new MainClass()) {
				if (!mainClass.InitializeGraphics()) {
					MessageBox.Show("Error while initializing Direct3D");
					return;
				}
				mainClass.Show();
				mainClass.Run();
			}
		}
	}
}
